#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
cf-configuration-exporter
"""

__program__ = "cf-configuration-exporter"
__version__ = "0.0.9"
__author__ = "Claudio Benfatto"
__year__ = "2017"
__email__ = "<claudio.benfatto@springer.com>"
__license__ = "MIT"



from jinja2 import Environment, PackageLoader, select_autoescape
from exporter.exporter import ResourceParser
import json

env = Environment(
    loader=PackageLoader('test', 'templates'),
)


class MockResourceFetcher:

    def __init__(self, client):
        self._client = client
        self.responses = {}

    def register_response(self, url, response):
        self.responses[url] = response

    def get_entities(self, resource_url):
        body = self.responses[resource_url]
        return ResourceParser.extract_entities(body)

class BasicMock:

  def get_response(self, resource):
    return self.template.render(resource=resource)


class UserUAAAPIMock(BasicMock):

  def __init__(self):
    self.template = env.get_template('user_uaa.j2')

class UserAPIMock(BasicMock):

  def __init__(self):
    self.template = env.get_template('user_cf.j2')


class OrgUsersAPIMock(BasicMock):

  def __init__(self):
    self.template = env.get_template('org_users.j2')


class SpaceAPIMock(BasicMock):

  def __init__(self, config, fetcher):
    self.template = env.get_template('space.j2')
    self._fetcher = fetcher
    self._config = config
    self._security_groups = list()

  @property
  def config(self):
      return self._config
  
  def get_guid(self):
    return self._config['guid']

  def dump(self):
      self.dump_sec_groups()
      space = json.loads(self.get_response(self._config))
      self._fetcher.register_response("/v2/spaces/%s" % self.get_guid(), space)
      return space

  def add_sec_group(self, sec_group):
    self._security_groups.append(sec_group.config)

  def dump_sec_groups(self):
    mock_sec_groups = SecGroupsAPIMock()
    sec_groups_definition = json.loads(mock_sec_groups.get_response(self._security_groups))
    self._fetcher.register_response("/v2/spaces/%s/security_groups" % self.get_guid(), 
        sec_groups_definition)


class OrganizationAPIMock(BasicMock):

  def __init__(self, config, fetcher):
    self.template = env.get_template('organization.j2')
    self._fetcher = fetcher
    self._config = config
    self._spaces = list()
    self._domains = list()

  @property
  def spaces(self):
      return self._spaces

  def get_guid(self):
    return self._config['guid']

  def dump(self):
      self.dump_spaces()
      self.dump_domains()

      org = json.loads(self.get_response(self._config))
      self._fetcher.register_response("/v2/organizations/%s" % self.get_guid(), org)
      return org

  def dump_spaces(self):
    mock_spaces = OrgSpacesAPIMock()
    spaces_definition = json.loads(mock_spaces.get_response(self._spaces))
    self._fetcher.register_response("/v2/organizations/%s/spaces" % self.get_guid(), 
                                      spaces_definition)

  def dump_domains(self):
    mock_domains = PrivateDomainsAPIMock()
    domains_definition = json.loads(mock_domains.get_response(self._domains))

    self._fetcher.register_response("/v2/organizations/%s/private_domains" % self.get_guid(), 
                                      domains_definition)

  def add_space(self, space):
    self._spaces.append(space.config)

  def add_domain(self, domain):
    self._domains.append(domain.config)

  def add_private_domains(self, private_domains):
    mock_private_domains = PrivateDomainsAPIMock()
    pd_definition = json.loads(mock_private_domains.get_response(private_domains))
    self._fetcher.register_response("/v2/organizations/%s/private_domains" % self.get_guid(), pd_definition)

  def _add_users(self, users, kind):

    mock_users = OrgUsersAPIMock()
    users_definition = json.loads(mock_users.get_response(users))
    self._fetcher.register_response("/v2/organizations/%s/%s" % (self.get_guid(), kind), 
                                          users_definition
                                      )

  def add_auditors(self, auditors):
    self._add_users(auditors, "auditors")

  def add_managers(self, managers):
    self._add_users(managers, "managers")

  def add_billing_managers(self, billing_managers):
    self._add_users(billing_managers, "billing_managers")

  def add_users(self, users):
    self._add_users(users, "users")


class QuotaAPIMock(BasicMock):

  def __init__(self, config, fetcher):
    self.template = env.get_template('quota.j2')
    self._config = config
    self._fetcher = fetcher

  def get_guid(self):
    return self._config['guid']

  def dump(self):
    quota = json.loads(self.get_response(self._config))
    self._fetcher.register_response("/v2/quota_definitions/%s" % self.get_guid(), quota)
    return quota


class SecGroupAPIMock(BasicMock):

  def __init__(self, config, fetcher):
    self.template = env.get_template('security_group.j2')
    self._config = config
    self._fetcher = fetcher

  @property
  def config(self):
      return self._config

class PrivateDomainAPIMock(BasicMock):

  def __init__(self, config, fetcher):
    self.template = env.get_template('domain.j2')
    self._config = config
    self._fetcher = fetcher

  @property
  def config(self):
      return self._config


class SecGroupsAPIMock:

  def __init__(self):
    self.template = env.get_template('security_groups.j2')

  def get_response(self, resource):
    return self.template.render(resource=resource)


class OrgSpacesAPIMock:

  def __init__(self):
    self.template = env.get_template('org_spaces.j2')

  def get_response(self, resource):
    return self.template.render(resource=resource)

class PrivateDomainsAPIMock:

  def __init__(self):
    self.template = env.get_template('private_domains.j2')

  def get_response(self, resource):
    return self.template.render(resource=resource)




import sys
import pyaml
import logging

from jinja2 import Environment, PackageLoader
from . import config as cfg
from .exporter import Exporter
from cfconfigurator.cf import CF

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger(__name__)

env = Environment(
    loader=PackageLoader('exporter', 'templates'),
)


def main():

	if cfg.api_url is None:
		logger.critical("Please set EXPORTER_API_URL env variable")
		sys.exit(1)
	if cfg.admin_user is None:
		logger.critical("Please set EXPORTER_ADMIN_USER env variable")
		sys.exit(1)
	if cfg.admin_password is None:
		logger.critical("Please set EXPORTER_ADMIN_PASSWORD env variable")
		sys.exit(1)

	logger.info("Start exporting configuration...")
	logger.info("CF api endpoint set to %s" % cfg.api_url)
	logger.info("Excluding the following env variables from the manifest: %s" % cfg.exclude_env_vars)

	cf_client = CF(cfg.api_url)
	cf_client.login(cfg.admin_user, cfg.admin_password)

	exp = Exporter(cf_client, exclude_vars=cfg.exclude_env_vars)
	exp.generate_manifest()

	manifest = {key: pyaml.dump({key:value}) for key, value in exp.manifest.items()}

	template = env.get_template('manifest.j2')
	rendered = template.render(manifest=manifest)


	with open(cfg.output_file ,"w") as stream:
		stream.write(rendered)

	logger.info("Manifest exported to '%s' file..." % (cfg.output_file))





from cfconfigurator.cf import CFException
from cfconfigurator.uaa import UAAException
from .exceptions import ExporterException 

import re
import sys
import logging
import collections
import functools

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger(__name__)


class ResourceParser(object):

    @classmethod
    def extract_entities(cls, body):
        if 'resources' in body:
            return [obj['entity'] for obj in body['resources']]
        else:
            return body['entity']

    @classmethod
    def extract_resources(cls, body):
        if 'resources' in body:
            return [obj for obj in body['resources']]

    @classmethod
    def extract_metadata(cls, body):
        return [obj['metadata'] for obj in body['resources']]


class Memoize(object):

    def __init__(self, make_request):
        self.make_request = make_request
        self.memo = {}

    def __call__(self, o_self, resource_url):
        if resource_url not in self.memo:
            response = self.make_request(o_self, resource_url)

            if response[1] == 200:
                self.memo[resource_url] = response
            else:
                return response

        return self.memo[resource_url]

    def __get__(self, instance, owner):
        from functools import partial
        return partial(self.__call__, instance)

class ResourceFetcher(object):
    """
    @brief      Help fetching resources from Cloudfoundry
    """

    def __init__(self, client):
        self._client = client


    def get_raw(self, resource_url):
        response = self.response(resource_url)
        return response[0]

    def get_resources(self, resource_url):
        """
        @brief      Get the `resources` configuration from the response body
        """
        response = self.response(resource_url)
        body = response[0]
        return ResourceParser.extract_resources(body)

    def get_entities(self, resource_url):
        """
        @brief      Get the `entities` configuration from the response body
        """
        response = self.response(resource_url)
        body = response[0]
        return ResourceParser.extract_entities(body)

    def get_metadata(self, resource_url):
        """
        @brief      Get the `metadata` configuration from the response body
        """
        response = self.response(resource_url)
        body = response[0]
        return ResourceParser.extract_metadata(body)

    @Memoize
    def response(self, resource_url):
        try:
            return self._client.request("GET", resource_url)
        except CFException as cfe:
            logger.err(str(cfe))
            raise

class BaseResource(object):
    """
    @brief      Base Class for a generic CF Resource
    """

    def __init__(self, *config_dicts, **kwargs):
        """
        @brief      Constructs the Resource object.
        
        @param      self          The object
        @param      config_dicts  The configuration dicts
        @param      fetcher       The fetcher
        """
        self._prop_dict = collections.OrderedDict()
        self._fetcher  = kwargs.get('fetcher', None)
        self._config = config_dicts

    def lookup(self, name):
        """
        @brief         looks up the name variable

        Performs a look up for a variable named `name` against
        the list of dictionaries passed as argument to the constructor
        of this class
        
        @param      self  The object
        @param      name  The name of the variable to look-up
        """
        for config in self._config:
            if name in config:
                return config[name] 
        raise AttributeError("%s not found" % name)

    def __getattr__(self, name):
        """
        Try to find the variable in the list of dictionaries passed to this
        instance contructor begore raising a AttributeError
        """
        return self.lookup(name)

    def load(self):
        """
        Parse the configuration and load the variables into the
        self._prop_dict instance dictionary
        """
        for prop in self.properties:
            try:
                value = getattr(self, prop)
                self._prop_dict[prop] = value
            except AttributeError as ate:
                pass

    def asdict(self):
        """
        @return     self._prop_dict instance dictionary containing a map representing
        the configuration parameters for this resource
        """
        return self._prop_dict

class FeatureFlag(BaseResource):
    """
    @brief:     Describe a feature flag.

    https://docs.cloudfoundry.org/adminguide/listing-feature-flags.html
    """
    properties = [  "name", 
                            "value"
                        ]

    @property
    def value(self):
        return self.__getattr__("enabled")

    def __init__(self, *config_dicts):
        super(FeatureFlag, self).__init__(*config_dicts)

class Vars(BaseResource):
    """
    @brief:     Describe environment variables groups.

    This resource in represented as a dictionary of
    `name`: `value` pairs where name is the name of the variable
    and value its value
    """
    properties = []

    def __init__(self, *config_dicts, **kwargs):
        super(Vars, self).__init__(*config_dicts)
        self.exclude_vars = kwargs.get('exclude_vars', ())


    def var_excluded(self, name):
        if name.lower().startswith(self.exclude_vars):
            return True
        return False

    def asdict(self):
        #Differetly from other resources we do not need
        #to extract the properties here. This resource consists
        #on just a dictionary of key: value entries
        return self._config[0]

    def aslist(self):
        return [{'name': name, 'value': value} for name, value in self._config[0].items() if not self.var_excluded(name)]

    def load(self):
        #Nothing to load. The response body is already in the correct format
        pass

class Quota(BaseResource):
    """
    @brief:     Describe a quota configuration

    Docs: https://docs.cloudfoundry.org/adminguide/quota-plans.html
     """

    properties = [  "name", 
                            "total_services", 
                            "total_routes", 
                            "memory_limit", 
                            "non_basic_services_allowed",
                            "instance_memory_limit", 
                            "total_service_keys",
                            "total_reserved_route_ports", 
                            "total_private_domains",
                            "app_instance_limit"
                        ]

    def __init__(self, *config_dicts):
        super(Quota, self).__init__(*config_dicts)

class Space(BaseResource):
    """
    @brief:     Describe a space configuration

    Docs: https://docs.cloudfoundry.org/concepts/roles.html#spaces
    """
    user_types = [
                            "developers", 
                            "managers", 
                            "auditors"
                        ]

    properties = [
                            "name", 
                            "allow_ssh", 
                            "developers", 
                            "managers", 
                            "auditors", 
                            "security_groups"
                        ]
    
    def __init__(self, *config_dicts,  **kwargs):
        super(Space, self).__init__(*config_dicts, **kwargs)
        self._security_groups = []


    @property
    def security_groups(self):
        return self._security_groups

    def load_security_groups(self):
        """
        @brief      extract and parse the security groups
        """
        url = self.lookup("security_groups_url")
        groups = self._fetcher.get_entities(url)
        if groups is None:
            return

        group_names = [group['name'] for group in groups if group['running_default'] is False]
        #at this point the group_names contain all the running groups in addition
        #to the groups assigned to this space.
        #That's why we need to remove the duplicates
        group_names = list(set(group_names))

        for name in group_names:
            self._security_groups.append({'name': name})

    def load_users(self):
        """
        @brief      extract and parse the users for this space
        """
        for user_type in self.user_types:
            url = "%s_url" % user_type
            try:
                users = self.lookup(url)
            except AttributeError as ate:
                continue
            user_list = []
            for user in users:
                if 'username' in user:
                    user_list.append(user['username'])
            if len(user_list) > 0:
                setattr(self, user_type, user_list)

    def load(self):
        self.load_users()
        self.load_security_groups()
        BaseResource.load(self)


class Organization(BaseResource):
    """
    @brief      Describe a CF organization
    """
    user_types = [
                            "users", 
                            "managers", 
                            "billing_managers", 
                            "auditors"
                        ]

    properties = [
                            "name", 
                            "quota",
                            "domains_private",
                            "spaces", 
                            "users", 
                            "managers", 
                            "billing_managers", 
                            "auditors"
                        ]

    def __init__(self, *config_dicts, **kwargs):
        super(Organization, self).__init__(*config_dicts, **kwargs)
        self._spaces = []

    @property
    def spaces(self):
        return self._spaces

    @property
    def quota(self):
        return self._quota

    @property
    def domains_private(self):
        return self._domains_private

    def load_private_domains(self):
        """
        @brief  Loads private domains list for this org
        """
        url = self.lookup("private_domains_url")
        p_domains = self._fetcher.get_entities(url)
        domains = []
        for domain in p_domains:
            if 'name' in domain:
                domains.append({'name': domain['name']})
        self._domains_private = domains

    def load_quota_definitions(self):
        """
        @brief      Loads quota definitions for this org.
        """
        url = self.lookup("quota_definition_url")
        quota = self._fetcher.get_entities(url)
        if 'name' in quota:
            self._quota = quota['name']

    def load_spaces(self):
        """
        @brief      Loads all the spaces for this org.
        """
        url = self.lookup("spaces_url")
        spaces = self._fetcher.get_entities(url)
        for space in spaces:
            new_space = Space(space, fetcher=self._fetcher)
            new_space.load()
            self._spaces.append(new_space.asdict())
    
    def load_users(self):
        """
        @brief      Loads all the users for this org.
        """
        for user_type in self.user_types:
            url = "%s_url" % user_type
            try:
                self.lookup(url)
            except AttributeError:
                continue 
            users = self._fetcher.get_entities(self.lookup(url))
            user_list = []
            for user in users:
                if 'username' in user:
                    user_list.append({'name': user['username']})
            if len(user_list) > 0:
                setattr(self, user_type, user_list)


    def load(self):
        self.load_private_domains()
        self.load_quota_definitions()
        self.load_spaces()
        self.load_users()
        BaseResource.load(self)

class SecurityGroup(BaseResource):
    """
    @brief      Describe a global security group
    """

    properties = [
                            "name", 
                            "running_default",
                            "staging_default", 
                            "rules"
                        ]

    def __init__(self, *config_dicts, **kwargs):
        super(SecurityGroup, self).__init__(*config_dicts, **kwargs)
        self._rules = []

    @property
    def rules(self):
        return self._rules 

    def generate_rule_name(self):
        num = 0
        while True:
            yield "Unnamed rule %i" % num
            num += 1

    def load(self):
        self.load_rules()
        BaseResource.load(self)

    def load_rules(self):
        rules = self.lookup('rules')
        name_generator = self.generate_rule_name()
        for index, rule in enumerate(rules):
            new_rule = SecurityRule(rule)
            new_rule.name_generator = name_generator
            if new_rule.has_name():
                new_rule.load()
                self._rules.append(new_rule.asdict())

class SecurityRule(BaseResource):
    """
    @brief:     Describe a security rule.
    """

    properties = [
                            "name", 
                            "protocol", 
                            "destination", 
                            "ports", 
                            "logs", 
                            "code", 
                            "type"
                        ]


    def __init__(self, *config_dicts, **kwargs):
        super(SecurityRule, self).__init__(*config_dicts, **kwargs)
        self._name_generator = None

    def has_name(self):
        try:
            self.lookup("description")
        except AttributeError:
            return False
        return True
    
    @property
    def name(self):
        return self.lookup("description")

    @property
    def name_generator(self):
        return self._name_generator

    @name_generator.setter
    def name_generator(self, value):
        self._name_generator = value


class User(BaseResource):
    """
    @brief      Describe a CF user.
    """

    properties = [
                            "name",
                            "password",
                            "active", 
                            "email", 
                            "given_name", 
                            "family_name",
                            "default_organization", 
                            "default_space", 
                            "origin", 
                            "external_id"
                        ]

    def __init__(self, *config_dicts, **kwargs):
        cf_response = kwargs.get('cf_response', None)
        fetcher  = kwargs.get('fetcher', None)

        if cf_response is None:
            raise ExporterException("Please provide the CF configuration for this user resource")
        if fetcher is None:
            raise ExporterException("Please provide a valid resource fetcher")

        super(User, self).__init__(*config_dicts, **kwargs)
        self._cf_response = cf_response

    def lookup_cf_response(self, name):
        config = self._cf_response
        if name in config:
                return config[name] 
        raise AttributeError("%s not found" % name)

    @property
    def external_id(self):
        return self.lookup("externalId")

    @property
    def given_name(self):
        name = self.lookup('name')
        if 'givenName' in name:
            return name['givenName']
        return None

    @property
    def family_name(self):
        name = self.lookup('name')
        if 'familyName' in name:
            return name['familyName']
        return None

    @property
    def password(self):
        return None
    
    @property
    def email(self):
        emails = self.lookup('emails')
        if len(emails) > 0:
            if 'value' in emails[0]:
                return emails[0]['value']
        raise AttributeError("email not found")
    
    @property
    def name(self):
        return self.lookup('userName')

    @property
    def default_space(self):
        return self._default_space

    @property
    def default_organization(self):
        return self._default_organization

    def load(self):
        self.load_default_space_and_org()
        BaseResource.load(self)

    def load_default_space_and_org(self):
        """
        @brief      Loads a default space and organization.
        """
        try:
            space_url = self.lookup_cf_response("default_space_url")
        except AttributeError:
            return
        space = self._fetcher.get_entities(space_url)
        if 'organization_url' in space:
            org_url = space['organization_url']
            org = self._fetcher.get_entities(org_url)

            if ('name' in space) and ('name' in org):
                self._default_organization = org['name']
                self._default_space = space['name']




class Exporter:

    def __init__(self, client, exclude_vars=""):
        self._client = client
        self._uaa_client = client.uaa
        self.fetcher = ResourceFetcher(client)
        self.manifest = collections.OrderedDict()
        exprs = re.split(';|,', exclude_vars)
        self.exclude_vars = tuple(expr.strip().lower() for expr in exprs if len(expr) > 0)

    def generate_manifest(self):
        self.manifest["cf_feature_flags"] = self.add_feature_flags()
        self.manifest["cf_staging_environment_variables"] = self.add_staging_environment_variables()
        self.manifest["cf_running_environment_variables"] = self.add_running_environment_variables()
        self.manifest["cf_shared_domains"] = self.add_shared_domains()
        self.manifest["cf_security_groups"] = self.add_security_groups()
        self.manifest["cf_quotas"] = self.add_quotas()
        self.manifest["cf_users"] = self.add_users()
        self.manifest["cf_orgs"] = self.add_orgs()

    def add_feature_flags(self):
        response = self.fetcher.get_raw("/v2/config/feature_flags")
        flag_list = []
        for flag in response:
            f = FeatureFlag(flag)
            f.load()
            flag_list.append(f.asdict())
        return flag_list

    def add_staging_environment_variables(self):
        response = self.fetcher.get_raw("/v2/config/environment_variable_groups/staging")
        
        v = Vars(response, exclude_vars=self.exclude_vars)
        return v.aslist()

    def add_running_environment_variables(self):
        response = self.fetcher.get_raw("/v2/config/environment_variable_groups/running")
        
        v = Vars(response, exclude_vars=self.exclude_vars)
        return v.aslist()

    def add_shared_domains(self):
        response = self.fetcher.get_entities("/v2/shared_domains")
        domain_list = []
        for domain in response:
            domain_list.append({'name': domain['name']})
        return domain_list

    def add_security_groups(self):
            response = self.fetcher.get_entities("/v2/security_groups")
            group_list = []
            for group in response:
                g = SecurityGroup(group)
                g.load()
                group_list.append(g.asdict())
            return group_list

    def add_quotas(self):
        response = self.fetcher.get_entities("/v2/quota_definitions")
        quota_list = []
        for quota in response:
            q = Quota(quota)
            q.load()
            quota_list.append(q.asdict())
        return quota_list

    def add_users(self):
        response = self.fetcher.get_resources("/v2/users")
        user_list = []
        for user in response:
            try:
                user_uaa = self._uaa_client.user_get(user['metadata']['guid'])
            except UAAException as uaaexp:
                continue
            user_cf = user['entity']
            u = User(user_uaa, cf_response=user_cf, fetcher=self.fetcher)
            u.load()
            user_list.append(u.asdict())
        return user_list

    def add_orgs(self):
        response = self.fetcher.get_entities("/v2/organizations")
        org_list = []
        for org in response:
            o = Organization(org, fetcher=self.fetcher)
            o.load()
            org_list.append(o.asdict())
        return org_list




#Configuration variables
import os

api_url = os.environ.get("EXPORTER_API_URL", None)
admin_user = os.environ.get("EXPORTER_ADMIN_USER", None)
admin_password = os.environ.get("EXPORTER_ADMIN_PASSWORD", None)
output_file = os.environ.get("EXPORTER_OUTPUT_FILE", "output")
exclude_env_vars = os.environ.get("EXPORTER_EXCLUDE_ENV_VARS", "")